"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@code-engine/utils");
const events_1 = require("events");
const ono_1 = require("ono");
const os = require("os");
const worker_1 = require("./worker");
/**
 * Runs CodeEngine plugins on worker threads.
 */
class WorkerPool extends events_1.EventEmitter {
    constructor(concurrency, context) {
        super();
        /** @internal */
        this._workers = [];
        /** @internal */
        this._isDisposed = false;
        /** @internal */
        this._moduleCounter = 0;
        /** @internal */
        this._roundRobinCounter = 0;
        concurrency = utils_1.validate.positiveInteger(concurrency, "concurrency", os.cpus().length);
        if (!context || typeof context.cwd !== "string") {
            throw ono_1.ono(`A CodeEngine context object is required.`);
        }
        this._createWorkers(concurrency, context);
    }
    /**
     * Indicates the number of worker threads in the pool.
     */
    get size() {
        return this._workers.length;
    }
    /**
     * Indicates whether the `dispose()` method has been called.
     * Once disposed, the `WorkerPool` instance is no longer usable.
     */
    get isDisposed() {
        return this._isDisposed;
    }
    /**
     * Loads the specified JavaScript module in all worker threads and returns a `FileProcessor` wrapper
     * that executes the module in one of the threads.
     */
    async loadFileProcessor(module) {
        this._assertNotDisposed();
        // Load the JavaScript module in all worker threads
        let [moduleUID, name] = await this._loadModule(module);
        // Create a CodeEngine FileProcessor function that executes the module on a worker thread
        let plugin = {
            [name]: (file, context) => {
                // Select a worker from the pool to process the files
                let worker = this._select();
                // Process the file on the worker thread
                return worker.processFile(moduleUID, file, context);
            }
        };
        // Return the FileProcessor function with the same name as the one in the module
        return plugin[name];
    }
    /**
     * Terminates all worker threads.
     */
    async dispose() {
        this._isDisposed = true;
        let workers = this._workers;
        this._workers = [];
        await Promise.all(workers.map((worker) => worker.terminate()));
    }
    /**
     * Creates the specified number of worker threads.
     * @internal
     */
    _createWorkers(concurrency, context) {
        // Re-emit all errros from workers
        let emitError = (error) => this.emit("error" /* Error */, error);
        for (let i = 0; i < concurrency; i++) {
            let worker = new worker_1.Worker(context);
            worker.on("error", emitError);
            this._workers.push(worker);
        }
    }
    /**
     * Loads the specified JavaScript module into all worker threads.
     * @internal
     */
    async _loadModule(module) {
        if (typeof module === "string") {
            module = { moduleId: module };
        }
        // Create a unique ID that will be used to reference this module from now on.
        let moduleUID = ++this._moduleCounter;
        let [name] = await Promise.all(this._workers.map((worker) => worker.loadModule(module, moduleUID)));
        return [moduleUID, name];
    }
    /**
     * Selects a `Worker` from the pool to perform a task.
     * @internal
     */
    _select() {
        // For now, we just use a simple round-robin strategy,
        // but we may employ a more advanced selection strategy later
        return this._workers[this._roundRobinCounter++ % this._workers.length];
    }
    /**
     * Throws an error if the `WorkerPool` has been disposed.
     * @internal
     */
    _assertNotDisposed() {
        if (this.isDisposed) {
            throw ono_1.ono(`CodeEngine cannot be used after it has been disposed.`);
        }
    }
}
exports.WorkerPool = WorkerPool;
//# sourceMappingURL=worker-pool.js.map